/*
    LBA LED flasher
    Copyright (C) 2011  Stefan Ott

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <error.h>
#include <errno.h>
#include <string.h>
#include <stdarg.h>

#include <arpa/inet.h>

#include "config.h"
#include "lba_blinkm.h"
#include "lba_status.h"

struct options {
	int port;
	struct in_addr host;
};

static void dbg_print(const char *msg, ...)
{
	va_list argp;

	va_start(argp, msg);
	vprintf(msg, argp);
	va_end(argp);
	printf("\n");
}

static void dbg_none(const char *msg, ...) {}

void (*debug)(const char *, ...) = &dbg_none;

static void help()
{
	printf(	"Usage: lba-led [OPTION]\n"
		"\n"
		"lba-led version %s\n"
		"\n"
		"Options:\n"
		"  -h			Show this help\n"
		"  -f			Don't fork\n"
		"  -d			Run in debug mode (implies -f)\n"
		"  -H			IP of host to check (default: %s)\n"
		"  -P			Port to check (default: %d)\n"
		"\n", PACKAGE_VERSION, LBA_CHECK_HOST, LBA_CHECK_PORT);
}

static int run(int led, struct nl80211_state *nlstate, struct options *opt)
{
	int bt, wifi, sys, l_bt, l_wifi, l_sys;
	int c_wifi[4], c_bt[4], c_sys[2];
	struct sockaddr_in host_addr;

	c_wifi[WIFI_SCANNING] = LBA_GREEN;
	c_wifi[WIFI_UP] = LBA_WHITE;
	c_wifi[WIFI_DOWN] = LBA_YELLOW;
	c_wifi[WIFI_ERROR] = LBA_RED;

	c_bt[BT_SCANNING] = LBA_BLUE;
	c_bt[BT_UP] = LBA_WHITE;
	c_bt[BT_DOWN] = LBA_YELLOW;
	c_bt[BT_ERROR] = LBA_RED;

	c_sys[SYS_LINK_UP] = LBA_GREEN;
	c_sys[SYS_LINK_DOWN] = LBA_RED;

	memcpy(&host_addr.sin_addr.s_addr, &opt->host, sizeof(struct in_addr));
	host_addr.sin_port = htons(opt->port);

	struct sockaddr *host = (struct sockaddr *) &host_addr;
	host->sa_family = AF_INET;

	while (1)
	{
		debug("Starting cycle");
		bt = btstatus();
		wifi = wifistatus(nlstate);
		sys = sysstatus(host);

		l_bt = c_bt[bt];
		l_wifi = c_wifi[wifi];
		l_sys = c_sys[sys];

		show_status(led, l_sys, l_wifi, l_bt);

		usleep(LBA_BLINK_SLEEP);
	}

	return EXIT_SUCCESS;
}

int main(int argc, char **argv)
{
	int c, do_fork, led, err;
	struct nl80211_state nlstate;
	char *ip;

	struct options opt;

	memset(&opt, 0, sizeof(struct options));

	do_fork = 1;
	ip = LBA_CHECK_HOST;
	opt.port = LBA_CHECK_PORT;

	while ((c = getopt(argc, argv, "dfhH:P:")) != -1)
	{
		switch(c)
		{
			case 'd':
				do_fork = 0;
				debug = &dbg_print;
				break;
			case 'f':
				do_fork = 0;
				break;
			case 'h':
				help();
				return EXIT_SUCCESS;
			case 'P':
				opt.port = atoi(optarg);
				if (opt.port < 1)
				{
					error(EXIT_FAILURE, EINVAL, optarg);
				}
				break;
			case 'H':
				ip = optarg;
				break;
			default:
				help();
				return EXIT_FAILURE;
		}
	}

	if (!inet_aton(ip, &opt.host))
	{
		error(EXIT_FAILURE, EINVAL, ip);
	}

	debug("Finished parsing arguments");
	debug("Using check host: %s:%d", inet_ntoa(opt.host), opt.port);

	led = init_blinkm();
	if (led < 0)
		return EXIT_FAILURE;
	
	err = -nl_init(&nlstate);

	if (err)
		error(EXIT_FAILURE, err, "Failed to initialize netlink socket");

	if (do_fork)
	{
		int pid_fork = fork();

		if (pid_fork < 0)
		{
			perror("Could not fork");
		}
		else if (pid_fork > 0)
		{
			return EXIT_SUCCESS;
		}
	}

	return run(led, &nlstate, &opt);
}
