/*
    LBA LED flasher
    Copyright (C) 2011  Stefan Ott

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#define _GNU_SOURCE

extern void (*debug)(const char *, ...);

#include <unistd.h>
#include <dirent.h>
#include <string.h>
#include <netdb.h>
#include <errno.h>

#include <sys/ioctl.h>

#include <linux/nl80211.h>

#include <netlink/msg.h>
#include <netlink/genl/genl.h>
#include <netlink/genl/family.h>
#include <netlink/genl/ctrl.h>

#include <bluetooth/bluetooth.h>
#include <bluetooth/hci.h>
#include <bluetooth/hci_lib.h>

#include "lba_status.h"

#define PROC_BASE "/proc"

#define COMM_LEN 16 	// synchronize with size of comm in struct task_struct
			// in /usr/include/linux/sched.h

int find_proc(char *name)
{
	DIR *dir;
	struct dirent *de;
	FILE *file;
	pid_t pid;
	char *path, comm[COMM_LEN];
	int hits = 0;

	if (!(dir = opendir(PROC_BASE)))
	{
		perror(PROC_BASE);
		return 0;
	}

	while ((de = readdir(dir)) != NULL)
	{
		if (!(pid = (pid_t) atoi(de->d_name)))
			continue;

		if (asprintf(&path, PROC_BASE "/%s/stat", de->d_name) < 0)
			continue;

		if (!(file = fopen(path, "r")))
		{
			free(path);
			continue;
		}
		free(path);

		if (fscanf(file, "%*d (%15[^)]", comm) != 1)
		{
			fclose(file);
			continue;
		}

		if (!strcmp(comm, name))
			hits++;

		fclose(file);
	}

	closedir(dir);

	debug("* Found %d running instance%s of %s", hits,
		hits == 1 ? "" : "s", name);

	return hits;
}

static int check_host(const struct sockaddr *host)
{
	int fd;
	struct timeval tmout;

	memset(&tmout, 0, sizeof(struct timeval));
	tmout.tv_usec = 100000;

	fd = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (fd == -1)
	{
		perror("Failed to open socket");
		return 1;
	}

	if (setsockopt(fd, SOL_SOCKET, SO_SNDTIMEO, &tmout, sizeof(tmout))== -1)
	{
		close(fd);
		perror("Failed to set socket timeout");
		return 1;
	}

	int ret = connect(fd, host, sizeof(struct sockaddr));

	if (ret == -1)
	{
		close(fd);
		perror("Could not connect");
		debug("--> System has no connectivity");
		return 1;
	}

	close(fd);

	debug("--> System seems OK!");
	return 0;
}

int sysstatus(const struct sockaddr *host)
{
	debug("\n===== Checking system =====");

	if (check_host(host))
		return SYS_LINK_DOWN;
	
	return SYS_LINK_UP;
}

int btstatus()
{
	int ctl;

	debug("\n===== Checking bluetooth =====");

	if ((ctl = socket(AF_BLUETOOTH, SOCK_RAW, BTPROTO_HCI)) < 0)
	{
		perror("Cannot open HCI socket");
		return BT_ERROR;
	}

	struct hci_dev_list_req *dl;
	struct hci_dev_req *dr;
	struct hci_dev_info di;
	int i, dev_up = 0;

	if (!(dl = malloc(HCI_MAX_DEV * sizeof(struct hci_dev_req) +
		sizeof(uint16_t))))
	{
		perror("Cannot allocate memory");
		close(ctl);
		return BT_ERROR;
	}

	dl->dev_num = HCI_MAX_DEV;
	dr = dl->dev_req;

	if (ioctl(ctl, HCIGETDEVLIST, (void *) dl) < 0)
	{
		perror("Cannot get device list");
		free(dl);
		close(ctl);
		return BT_ERROR;
	}

	if (dl->dev_num < 1)
	{
		debug("* No bluetooth devices found");
		free(dl);
		close(ctl);
		return BT_ERROR;
	}
	debug("* Found %d bluetooth device(s)", dl->dev_num);

	for (i = 0; i< dl->dev_num; i++)
	{
		di.dev_id = (dr+i)->dev_id;
		if (ioctl(ctl, HCIGETDEVINFO, (void *) &di) < 0)
			continue;

		if (hci_test_bit(HCI_UP, &di.flags))
		{
			debug("*  Device hci%d is up", i);
			dev_up++;
		}
		else
		{
			debug("*  Device hci%d is down", i);
		}
	}

	if (dev_up < 1)
	{
		debug("--> All bluetooth devices are down");
		free(dl);
		close(ctl);
		return BT_DOWN;
	}

	debug("* Found %d total active bluetooth device%s", dev_up,
		dev_up == 1 ? "" : "s");

	free(dl);
	close(ctl);

	if (find_proc("lba-btscan") < 1)
	{
		debug("--> Bluetooth is up but not scanning");
		return BT_UP;
	}

	debug("--> Bluetooth is up & scanning");
	return BT_SCANNING;
}

static int count_iface_types(struct nl_msg *msg, void *arg)
{
	struct nlattr *tb_msg[NL80211_ATTR_MAX + 1];
	struct genlmsghdr *gnlh = nlmsg_data(nlmsg_hdr(msg));
	enum nl80211_iftype type;
	struct iface_count *count = (struct iface_count *) arg;

	nla_parse(tb_msg, NL80211_ATTR_MAX, genlmsg_attrdata(gnlh, 0),
		genlmsg_attrlen(gnlh, 0), NULL);

	if (tb_msg[NL80211_ATTR_IFTYPE])
	{
		type = nla_get_u32(tb_msg[NL80211_ATTR_IFTYPE]);

		if (type == NL80211_IFTYPE_MONITOR)
		{
			debug("* Interface %s is in monitor mode",
				nla_get_string(tb_msg[NL80211_ATTR_IFNAME]));
			count->NL80211_IFTYPE_MONITOR += 1;
		}
		else
		{
			debug("* Interface %s is in mode %d (non-monitor)",
				nla_get_string(tb_msg[NL80211_ATTR_IFNAME]),
				type);
			count->NL80211_IFTYPE_UNSPECIFIED += 1;
		}
	}

	return 0;
}

int wifistatus(struct nl80211_state *state)
{
	struct nl_cb *cb;
	struct nl_msg *msg;
	struct iface_count count;

	debug("\n===== Checking wifi =====");

	count.NL80211_IFTYPE_MONITOR = 0;
	count.NL80211_IFTYPE_UNSPECIFIED = 0;

	msg = nlmsg_alloc();

	if (!msg)
	{
		fprintf(stderr, "Failed to allocate netlink message\n");
		return WIFI_ERROR;
	}

	cb = nl_cb_alloc(NL_CB_DEFAULT);

	if (!cb)
	{
		fprintf(stderr, "Failed to allocate netlink callbacks\n");
		nl_socket_free(state->nl_sock);
		nlmsg_free(msg);
		return WIFI_ERROR;
	}

	if (nl_cb_set(cb, NL_CB_VALID, NL_CB_CUSTOM, count_iface_types, &count))
	{
		fprintf(stderr, "Failed to set up netlink callback\n");
	}

	// TODO check return values
	genlmsg_put(msg, 0, 0, genl_family_get_id(state->nl80211), 0,
		NLM_F_DUMP, NL80211_CMD_GET_INTERFACE, 0);

	nl_send_auto_complete(state->nl_sock, msg);
	nlmsg_free(msg);

	nl_recvmsgs(state->nl_sock, cb);
	// END TODO

	debug("* Found %d interface(s) in monitor mode and %d in other modes",
		count.NL80211_IFTYPE_MONITOR, count.NL80211_IFTYPE_UNSPECIFIED);

	if (count.NL80211_IFTYPE_MONITOR < 1)
	{
		if (count.NL80211_IFTYPE_UNSPECIFIED > 1)
		{
			debug("--> Wifi is up but not scanning");
			return WIFI_UP;
		}
		else
		{
			debug("--> Wifi is down");
			return WIFI_DOWN;
		}
	}

	// this should not happen if the scanner is started/stopped through
	// the init script, could happen if it dies unepectedly or is killed
	// manually though.
	if (find_proc("lba-wifiscan") < 1)
	{
		debug("--> Wifi is up but not scanning");
		return WIFI_UP;
	}

	debug("--> Wifi is up & scanning");
	return WIFI_SCANNING;
}

int nl_init(struct nl80211_state *state)
{
	state->nl_sock = nl_socket_alloc();

	if (!state->nl_sock)
	{
		fprintf(stderr, "Failed to allocate netlink socket\n");
		return -ENOMEM;
	}

	if (genl_connect(state->nl_sock))
	{
		fprintf(stderr, "Failed to connect to generic netlink.\n");
		nl_socket_free(state->nl_sock);
		return -ENOLINK;
	}

	if (genl_ctrl_alloc_cache(state->nl_sock, &state->nl_cache))
	{
		fprintf(stderr, "Failed to allocate generic netlink cache\n");
		nl_socket_free(state->nl_sock);
		return -ENOMEM;
	}

	state->nl80211 = genl_ctrl_search_by_name(state->nl_cache, "nl80211");
	if (!state->nl80211) {
		fprintf(stderr, "nl80211 not found.\n");
		nl_socket_free(state->nl_sock);
		nl_cache_free(state->nl_cache);
		return -ENOENT;
	}

	return 0;
}
